@@ -1,4 +1,4 @@
-use darling::FromField;
+use darling::{FromDeriveInput, FromField};
 use proc_macro::TokenStream;
 use quote::quote;
 use std::iter;
@@ -32,7 +32,7 @@ pub fn init(_: TokenStream, function: TokenStream) -> TokenStream {
             use super::*;
 
             #[no_mangle]
-            unsafe extern "C" fn init_plugin(plugin: *mut ::panda::PluginHandle) {
+            pub unsafe extern "C" fn init_plugin(plugin: *mut ::panda::PluginHandle) -> bool {
                 ::panda::set_plugin_ref(plugin);
 
                 for cb in ::panda::inventory::iter::<::panda::InternalCallback> {
@@ -43,11 +43,11 @@ pub fn init(_: TokenStream, function: TokenStream) -> TokenStream {
                     cb.0();
                 }
 
-                #func_name(#args);
+                ::panda::InitReturn::into_init_bool(#func_name(#args))
             }
 
             #[no_mangle]
-            unsafe extern "C" fn uninit_plugin(plugin: *mut ::panda::PluginHandle) {
+             pub unsafe extern "C" fn uninit_plugin(plugin: *mut ::panda::PluginHandle) {
                 for cb in ::panda::inventory::iter::<::panda::UninitCallback> {
                     cb.0(unsafe { &mut *plugin });
                 }
@@ -163,6 +163,18 @@ pub fn hook(_: TokenStream, func: TokenStream) -> TokenStream {
     .into()
 }
 
+mod guest_type;
+use guest_type::GuestTypeInput;
+
+#[proc_macro_derive(GuestType, attributes(name, arg))]
+pub fn derive_guest_type(input: TokenStream) -> TokenStream {
+    let input = syn::parse_macro_input!(input as syn::DeriveInput);
+    match GuestTypeInput::from_derive_input(&input) {
+        Ok(input) => input.to_tokens().into(),
+        Err(err) => err.write_errors().into(),
+    }
+}
+
 #[proc_macro_attribute]
 pub fn channel_recv(_: TokenStream, func: TokenStream) -> TokenStream {
     let mut func = syn::parse_macro_input!(func as syn::ItemFn);
@@ -195,151 +207,8 @@ pub fn channel_recv(_: TokenStream, func: TokenStream) -> TokenStream {
     .into()
 }
 
-#[derive(FromField)]
-#[darling(attributes(arg))]
-struct DeriveArgs {
-    #[darling(default)]
-    about: Option<String>,
-    #[darling(default)]
-    default: Option<syn::Lit>,
-    #[darling(default)]
-    required: bool,
-    ident: Option<syn::Ident>,
-    ty: syn::Type,
-}
-
-fn derive_args_to_mappings(
-    DeriveArgs {
-        about,
-        default,
-        ident,
-        ty,
-        required,
-    }: DeriveArgs,
-) -> (syn::Stmt, syn::Ident) {
-    let name = &ident;
-    let default = if let Some(default) = default {
-        match default {
-            syn::Lit::Str(string) => quote!(::std::string::String::from(#string)),
-            default => quote!(#default),
-        }
-    } else {
-        quote!(Default::default())
-    };
-    let about = about.unwrap_or_default();
-    (
-        syn::parse_quote!(
-            let #name = <#ty as ::panda::panda_arg::GetPandaArg>::get_panda_arg(
-                __args_ptr,
-                stringify!(#name),
-                #default,
-                #about,
-                #required
-            );
-        ),
-        ident.unwrap(),
-    )
-}
-
-fn get_field_statements(
-    fields: &syn::Fields,
-) -> Result<(Vec<syn::Stmt>, Vec<syn::Ident>), darling::Error> {
-    Ok(fields
-        .iter()
-        .map(DeriveArgs::from_field)
-        .collect::<Result<Vec<_>, _>>()?
-        .into_iter()
-        .map(derive_args_to_mappings)
-        .unzip())
-}
-
-fn get_name(attrs: &[syn::Attribute]) -> Option<String> {
-    attrs
-        .iter()
-        .find(|attr| attr.path.get_ident().map(|x| *x == "name").unwrap_or(false))
-        .map(|attr| attr.parse_meta().ok())
-        .flatten()
-        .map(|meta| {
-            if let syn::Meta::NameValue(syn::MetaNameValue {
-                lit: syn::Lit::Str(s),
-                ..
-            }) = meta
-            {
-                Some(s.value())
-            } else {
-                None
-            }
-        })
-        .flatten()
-}
-
-#[proc_macro_derive(PandaArgs, attributes(name, arg))]
-pub fn derive_panda_args(input: TokenStream) -> TokenStream {
-    let input = syn::parse_macro_input!(input as syn::ItemStruct);
-
-    let name = match get_name(&input.attrs) {
-        Some(name) => name,
-        None => {
-            return quote!(compile_error!(
-                "Missing plugin name, add `#[name = ...]` above struct"
-            ))
-            .into()
-        }
-    };
-
-    let ident = &input.ident;
-
-    match get_field_statements(&input.fields) {
-        Ok((statements, fields)) => {
-            let format_args = iter::repeat("{}={}")
-                .take(statements.len())
-                .collect::<Vec<_>>()
-                .join(",");
-            quote!(
-                impl ::panda::PandaArgs for #ident {
-                    const PLUGIN_NAME: &'static str = #name;
-
-                    fn from_panda_args() -> Self {
-                        let name = ::std::ffi::CString::new(#name).unwrap();
-
-                        unsafe {
-                            let __args_ptr = ::panda::sys::panda_get_args(name.as_ptr());
-
-                            #(
-                                #statements
-                            )*
-
-                            ::panda::sys::panda_free_args(__args_ptr);
-
-                            Self {
-                                #(#fields),*
-                            }
-                        }
-                    }
-
-                    fn to_panda_args_str(&self) -> ::std::string::String {
-                        format!(
-                            concat!(#name, ":", #format_args),
-                            #(
-                                stringify!(#fields), self.#fields
-                            ),*
-                       )
-                    }
-
-                    fn to_panda_args(&self) -> ::std::vec::Vec<(&'static str, ::std::string::String)> {
-                        ::std::vec![
-                            #(
-                                (stringify!(#fields), self.#fields.to_string()),
-                            )*
-                        ]
-                    }
-                }
-            )
-        }
-        Err(err) => err.write_errors(),
-    }
-    .into()
-}
+// derive PandaArgs
+include!("panda_args.rs");
 
 struct Idents(syn::Ident, syn::Ident);
 
@@ -419,7 +288,6 @@ macro_rules! define_callback_attributes {
                             }
 
                             pub fn disable() {
-                                println!("plugin: {}", env!("CARGO_PKG_NAME"));
                                 unsafe {
                                     ::panda::sys::panda_disable_callback(
                                         ::panda::sys::panda_get_plugin_by_name(
